#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import inspect
from typing import Any, Dict, Type
from .base import ModelTransformer


__all__ = ["get_model_transformer"]
_MODEL_TRANSFORMERS: Dict[type, Type[ModelTransformer]] = {}


def get_model_transformer(model: Any, **kwargs: Any) -> ModelTransformer:
    """Creates a ModelTransformer for a given model or raises an exception if one is not available"""
    for model_type, transformer in _MODEL_TRANSFORMERS.items():
        if isinstance(model, model_type):
            # Filter out kwargs that aren't applicable to the specific 'ModelTransformer'
            accepted_kwargs = {
                param for param in inspect.signature(transformer.__init__).parameters
            }
            kwargs = {k: v for k, v in kwargs.items() if k in accepted_kwargs}

            return transformer(model, **kwargs)

    raise NotImplementedError(
        f"Importing ML models of type {type(model)}, not currently implemented"
    )


try:
    from .sklearn import (
        SKLearnDecisionTreeTransformer,
        SKLearnForestClassifierTransformer,
        SKLearnForestRegressorTransformer,
        SKLearnForestTransformer,
        SKLearnTransformer,
        _MODEL_TRANSFORMERS as _SKLEARN_MODEL_TRANSFORMERS,
    )

    __all__ += [
        "SKLearnDecisionTreeTransformer",
        "SKLearnForestClassifierTransformer",
        "SKLearnForestRegressorTransformer",
        "SKLearnForestTransformer",
        "SKLearnTransformer",
    ]
    _MODEL_TRANSFORMERS.update(_SKLEARN_MODEL_TRANSFORMERS)
except ImportError:
    pass

try:
    from .xgboost import (
        XGBoostClassifierTransformer,
        XGBClassifier,
        XGBoostForestTransformer,
        XGBoostRegressorTransformer,
        XGBRegressor,
        _MODEL_TRANSFORMERS as _XGBOOST_MODEL_TRANSFORMERS,
    )

    __all__ += [
        "XGBoostClassifierTransformer",
        "XGBClassifier",
        "XGBoostForestTransformer",
        "XGBoostRegressorTransformer",
        "XGBRegressor",
    ]
    _MODEL_TRANSFORMERS.update(_XGBOOST_MODEL_TRANSFORMERS)
except ImportError:
    pass

try:
    from .lightgbm import (
        LGBMRegressor,
        LGBMClassifier,
        LGBMForestTransformer,
        LGBMRegressorTransformer,
        LGBMClassifierTransformer,
        _MODEL_TRANSFORMERS as _LIGHTGBM_MODEL_TRANSFORMERS,
    )

    __all__ += [
        "LGBMRegressor",
        "LGBMClassifier",
        "LGBMForestTransformer",
        "LGBMRegressorTransformer",
        "LGBMClassifierTransformer",
    ]
    _MODEL_TRANSFORMERS.update(_LIGHTGBM_MODEL_TRANSFORMERS)
except ImportError:
    pass



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# File called _pytest for PyCharm compatability
import eland as ed
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_INDEX_NAME
from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_series_equal


class TestSeriesHeadTail(TestData):
    def test_head_tail(self):
        pd_s = self.pd_flights()["Carrier"]
        ed_s = ed.Series(ES_TEST_CLIENT, FLIGHTS_INDEX_NAME, "Carrier")

        pd_s_head = pd_s.head(10)
        ed_s_head = ed_s.head(10)

        assert_pandas_eland_series_equal(pd_s_head, ed_s_head)

        pd_s_tail = pd_s.tail(10)
        ed_s_tail = ed_s.tail(10)

        assert_pandas_eland_series_equal(pd_s_tail, ed_s_tail)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# File called _pytest for PyCharm compatability

from pandas.testing import assert_series_equal

from eland.tests.common import TestData


class TestDataFrameNUnique(TestData):
    def test_flights_nunique(self):
        # Note pandas.nunique fails for dict columns (e.g. DestLocation)
        columns = [
            "AvgTicketPrice",
            "Cancelled",
            "Carrier",
            "Dest",
            "DestAirportID",
            "DestCityName",
        ]
        pd_flights = self.pd_flights()[columns]
        ed_flights = self.ed_flights()[columns]

        pd_flights.nunique()
        ed_flights.nunique()

        # TODO - ES is approximate counts so these aren't equal...
        # E[left]: [13059, 2, 4, 156, 156, 143]
        # E[right]: [13132, 2, 4, 156, 156, 143]
        # assert_series_equal(pd_nunique, ed_nunique)

    def test_ecommerce_nunique(self):
        columns = ["customer_first_name", "customer_last_name", "day_of_week_i"]
        pd_ecommerce = self.pd_ecommerce()[columns]
        ed_ecommerce = self.ed_ecommerce()[columns]

        pd_nunique = pd_ecommerce.nunique()
        ed_nunique = ed_ecommerce.nunique()

        assert_series_equal(pd_nunique, ed_nunique)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pandas as pd
from eland.tests.common import TestData, assert_series_equal


class TestSeriesDescribe(TestData):
    def test_series_describe(self):
        ed_df = self.ed_flights_small()
        pd_df = self.pd_flights_small()

        ed_desc = ed_df.AvgTicketPrice.describe()
        pd_desc = pd_df.AvgTicketPrice.describe()

        assert isinstance(ed_desc, pd.Series)
        assert ed_desc.shape == pd_desc.shape
        assert ed_desc.dtype == pd_desc.dtype
        assert ed_desc.index.equals(pd_desc.index)

        # Percentiles calculations vary for Elasticsearch
        assert_series_equal(
            ed_desc[["count", "mean", "std", "min", "max"]],
            pd_desc[["count", "mean", "std", "min", "max"]],
            rtol=0.2,
        )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from typing import Optional, TextIO, TYPE_CHECKING
from eland.utils import deprecated_api

if TYPE_CHECKING:
    from .query_compiler import QueryCompiler


class Index:
    """
    The index for an eland.DataFrame.

    TODO - This currently has very different behaviour than pandas.Index

    Currently, the index is a field that exists in every document in an Elasticsearch index.
    For slicing and sorting operations it must be a docvalues field. By default _id is used,
    which can't be used for range queries and is inefficient for sorting:

    https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-id-field.html
    (The value of the _id field is also accessible in aggregations or for sorting,
    but doing so is discouraged as it requires to load a lot of data in memory.
    In case sorting or aggregating on the _id field is required, it is advised to duplicate
    the content of the _id field in another field that has doc_values enabled.)
    """

    ID_INDEX_FIELD = "_id"
    ID_SORT_FIELD = "_doc"  # if index field is _id, sort by _doc

    def __init__(
        self, query_compiler: "QueryCompiler", es_index_field: Optional[str] = None
    ):
        self._query_compiler = query_compiler

        # _is_source_field is set immediately within
        # index_field.setter
        self._is_source_field = False

        # The type:ignore is due to mypy not being smart enough
        # to recognize the property.setter has a different type
        # than the property.getter.
        self.es_index_field = es_index_field  # type: ignore

    @property
    def sort_field(self) -> str:
        if self._index_field == self.ID_INDEX_FIELD:
            return self.ID_SORT_FIELD
        return self._index_field

    @property
    def is_source_field(self) -> bool:
        return self._is_source_field

    @property
    def es_index_field(self) -> str:
        return self._index_field

    @es_index_field.setter
    def es_index_field(self, index_field: Optional[str]) -> None:
        if index_field is None or index_field == Index.ID_INDEX_FIELD:
            self._index_field = Index.ID_INDEX_FIELD
            self._is_source_field = False
        else:
            self._index_field = index_field
            self._is_source_field = True

    def __len__(self) -> int:
        return self._query_compiler._index_count()

    # Make iterable
    def __next__(self) -> None:
        # TODO resolve this hack to make this 'iterable'
        raise StopIteration()

    def __iter__(self) -> "Index":
        return self

    def es_info(self, buf: TextIO) -> None:
        buf.write("Index:\n")
        buf.write(f" es_index_field: {self.es_index_field}\n")
        buf.write(f" is_source_field: {self.is_source_field}\n")

    @deprecated_api("eland.Index.es_info()")
    def info_es(self, buf: TextIO) -> None:
        self.es_info(buf)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# File called _pytest for PyCharm compatability

import pytest
from eland.tests.common import TestData
from eland.tests.common import assert_eland_frame_equal


class TestDataEsQuery(TestData):
    def test_flights_match_query(self):
        ed_flights = self.ed_flights()

        left = ed_flights.es_query({"match": {"OriginCityName": "Rome"}})[
            ed_flights["Carrier"] == "Kibana Airlines"
        ]

        right = ed_flights[ed_flights["Carrier"] == "Kibana Airlines"].es_query(
            {"match": {"OriginCityName": "Rome"}}
        )

        assert len(left) > 0
        assert_eland_frame_equal(left, right)

    def test_es_query_allows_query_in_dict(self):
        ed_flights = self.ed_flights()

        left = ed_flights.es_query({"match": {"OriginCityName": "Rome"}})
        right = ed_flights.es_query({"query": {"match": {"OriginCityName": "Rome"}}})

        assert len(left) > 0
        assert_eland_frame_equal(left, right)

    def test_es_query_geo_location(self):
        df = self.ed_ecommerce()
        cur_nearby = df.es_query(
            {
                "bool": {
                    "filter": {
                        "geo_distance": {
                            "distance": "10km",
                            "geoip.location": {"lon": 55.3, "lat": 25.3},
                        }
                    }
                }
            }
        )["currency"].value_counts()

        assert cur_nearby["EUR"] == 476

    @pytest.mark.parametrize("query", [(), [], 1, True])
    def test_es_query_wrong_type(self, query):
        ed_flights = self.ed_flights_small()

        with pytest.raises(TypeError):
            ed_flights.es_query(query)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# File called _pytest for PyCharm compatability

import pytest
from eland.tests.common import TestData


class TestDataFrameFilter(TestData):
    def test_filter_arguments_mutually_exclusive(self, df):
        with pytest.raises(TypeError):
            df.filter(items=[], like="!", regex="!")
        with pytest.raises(TypeError):
            df.filter(items=[], regex="!")
        with pytest.raises(TypeError):
            df.filter(items=[], like="!")
        with pytest.raises(TypeError):
            df.filter(like="!", regex="!")
        with pytest.raises(TypeError):
            df.filter()

    @pytest.mark.parametrize(
        "items",
        [
            ["DestCountry", "Cancelled", "AvgTicketPrice"],
            [],
            ["notfound", "AvgTicketPrice"],
        ],
    )
    def test_filter_columns_items(self, df, items):
        df.filter(items=items)

    @pytest.mark.parametrize("like", ["Flight", "Nope"])
    def test_filter_columns_like(self, df, like):
        df.filter(like=like)

    @pytest.mark.parametrize("regex", ["^Flig", "^Flight.*r$", ".*", "^[^C]"])
    def test_filter_columns_regex(self, df, regex):
        df.filter(regex=regex)

    @pytest.mark.parametrize("items", [[], ["20"], [str(x) for x in range(30)]])
    def test_filter_index_items(self, df, items):
        df.filter(items=items, axis=0)

    def test_filter_index_like_and_regex(self):
        ed_flights_small = self.ed_flights_small()

        with pytest.raises(NotImplementedError):
            ed_flights_small.filter(like="2", axis=0)
        with pytest.raises(NotImplementedError):
            ed_flights_small.filter(regex="^2", axis=0)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import re
import functools
import warnings
from typing import Callable, TypeVar, Any, Union, List, cast, Collection, Iterable
from collections.abc import Collection as ABCCollection
import pandas as pd  # type: ignore


RT = TypeVar("RT")
Item = TypeVar("Item")


def deprecated_api(
    replace_with: str,
) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    def wrapper(f: Callable[..., RT]) -> Callable[..., RT]:
        @functools.wraps(f)
        def wrapped(*args: Any, **kwargs: Any) -> RT:
            warnings.warn(
                f"{f.__name__} is deprecated, use {replace_with} instead",
                DeprecationWarning,
                stacklevel=2,
            )
            return f(*args, **kwargs)

        return wrapped

    return wrapper


def is_valid_attr_name(s: str) -> bool:
    """
    Ensure the given string can be used as attribute on an object instance.
    """
    return bool(
        isinstance(s, str) and re.search(string=s, pattern=r"^[a-zA-Z_][a-zA-Z0-9_]*$")
    )


def to_list(x: Union[Collection[Any], pd.Series]) -> List[Any]:
    if isinstance(x, ABCCollection):
        return list(x)
    elif isinstance(x, pd.Series):
        return cast(List[Any], x.to_list())
    raise NotImplementedError(f"Could not convert {type(x).__name__} into a list")


def try_sort(iterable: Iterable[Item]) -> Iterable[Item]:
    # Pulled from pandas.core.common since
    # it was deprecated and removed in 1.1
    listed = list(iterable)
    try:
        return sorted(listed)
    except TypeError:
        return listed
